// For use with Scilab 
// Copyright (C) 2017 -- Jérôme Mendes <jermendes@gmail.com> or <jermendes@isr.uc.pt>, Luís Osório <lbica@hotmail.com>, Rui Araújo <rui@isr.uc.pt>
// Jérôme Mendes, Luís Osório, Rui Araújo, "The self-tuning pid controllers toolbox", 2017.
// The "The self-tuning pid controllers toolbox" comes with ABSOLUTELY NO WARRANTY;
// In case of publication of any application of these methods,
// please, cite the work:
// Jérôme Mendes, Luís Osório, and Rui Araújo. 
// Self-tuning pid controllers in pursuit of plug and play capacity. 
// Control Engineering Practice, 69:73-84, December 2017.
// DOI: https://doi.org/10.1016/j.conengprac.2017.09.006
// http://www.isr.uc.pt/~jermendes/    ;     http://www.isr.uc.pt/~rui/
function [u,na,nb,theta,X,i,w,vx,I_mU,ye,ep,sig]=Tuning_And_PID(theta,T0,e,u,y,r,k,kk,X,i,w,vx,ye,ep,I_mU,sig,Max_u,Min_u,Max_u_var,method)
    
///////////////////////////////////////////////////////////////////////////////
//  Dahlin PID controller

    if method == 'Dahlin' | method == '1'  then

        // Coeficients for Dahlin PID
        B =200// Defined by the user    
        
        //Identified System 
        a1=theta(1,k-1);
        a2=theta(2,k-1);             
        b1=theta(3,k-1);
   
        Q= 1 - exp(-T0/B);
        
        Kp = -(a1 + 2*a2) * (Q/b1);
        Td = T0*a2*Q / (Kp*b1);
        Ti = - T0/((1 / (a1 + 2*a2)) + 1 + (Td / T0));

	// PID
        if k>3 then
            u(k-1)= Kp*( e(k-1)*(1+(T0/Ti)+(Td/T0)) - e(k-2)*(1+2*Td/T0) + e(k-3)*(Td/T0)) + u(k-2)
        else
            u(k-1) = Kp*( e(k-1)*(1+(T0/Ti)+(Td/T0)))
        end
    
///////////////////////////////////////////////////////////////////////////////
//  Pole placemente controller

    elseif method == 'Pole Placement' | method == '3'  then
        
        //Identified System 
        a1=theta(1,k-1);
        a2=theta(2,k-1);             
        b1=theta(3,k-1);
        b2=theta(4,k-1);
    
        xi =8; //the lower the more reactive // Defined by the user    
        om = 0.2/T0; // Defined by the user    
    
        d2=exp(-2*xi*om*T0);
        if xi <= 1 then
           d1=-2*exp(-xi*om*T0)*cos(om*T0*(sqrt(1-xi*xi)));
        else
           d1=-2*exp(-xi*om*T0)*cosh(om*T0*(sqrt(xi*xi-1)));
        end

        s1 = a2*((b1+b2)*(a1*b2-a2*b1)+b2*(b1*d2-b2*d1-b2));
        r1 = (b1+b2)*(a1*b1*b2-a2*b1*b1-b2*b2);
    
        q2 = s1/r1;
        gama = q2*b2/a2;
        q1 = (a2/b2)-q2*(b1/b2-a1/a2+1);
        q0 = (1/b1) * (d1+1-a1-gama);
        p1 = gama-1;
        p2 = -gama;
    
        // PID
        if k>4 then
            u(k-1) = e(k-1)*q0 + e(k-2)*q1 + e(k-3)*q2 - p1*u(k-2) - p2*u(k-3)
        else
            u(k-1) = e(k-1)*q0
        end
    
///////////////////////////////////////////////////////////////////////////////
//   Deadbeat controller for processes of secound order

    elseif method == 'Deadbeat 2nd' | method == '4' then
        
    	//Identified System 
        a1=theta(1,k-1);
        a2=theta(2,k-1);
        b1=theta(3,k-1);
        b2=theta(4,k-1);
        
        pq = inv([ 1 b1  0 ; a1 b2 b1 ; a2  0 b2]) * [-a1; -a2; 0];
        
        p1 = pq(1);
        q0 = pq(2);
        q1 = pq(3);
        r0 = 1/(b1+b2);
    
	// PID
        if k>3 then
            u(k-1) =  r0*r(k-1) - q0*y(k-1) - q1*y(k-2) - p1*u(k-2)
        else
            u(k-1) =  r0*r(k-1) - q0*y(k-1)
        end

///////////////////////////////////////////////////////////////////////////////
//  Deadbeat controllerfor processes of third order

    elseif method == 'Deadbeat 3rd' | method == '5' then
        
        //Identified System 
        a1=theta(1,k-1);
        a2=theta(2,k-1);
        a3=theta(3,k-1);
        b1=theta(4,k-1);
        b2=theta(5,k-1);
        b3=theta(6,k-1);
        
        pq =inv([1   0  b1   0   0;
                a1   1  b2  b1   0;
                a2  a1  b3  b2  b1; 
                a3  a2   0  b3  b2; 
                 0  a3   0   0  b3])*[-a1; -a2; -a3; 0; 0];
        p1 = pq(1);
        p2 = pq(2);
        q0 = pq(3);
        q1 = pq(4);
        q2 = pq(5);
  
        r0 = 1/(b1+b2+b3);
 	
	// PID
        if k>4 then
            u(k-1) =  r0*r(k-1) - q0*y(k-1) - q1*y(k-2) - q2*y(k-3) - p1*u(k-2) - p2*u(k-3)
        else
            u(k-1) =  r0*r(k-1) - q0*y(k-1) 
        end
      
        
///////////////////////////////////////////////////////////////////////////////
//  Ziegler-Nichols PID controller based on forward rectangular method of discretization for processes of third order

    elseif method == 'Ziegler Nichols FR' | method == '6' then 
    
    	//Identified System 
        a1=theta(1,k-1);
        a2=theta(2,k-1);
        a3=theta(3,k-1)
        b1=theta(4,k-1);
        b2=theta(5,k-1);
        b3=theta(6,k-1);
        
    // Determine Ultimate Gain and Ultimate Period
        
       K_c = [0 0];	//gain for complex roots // Defined by the user    
       aa = (b1-b3)*b3;
       bb = (b1-2*b3)*a3 + b3*a1 - b2;
       cc = a3*(a1-a3)-a2+1;
       if (aa ~= 0)
          D = bb*bb - 4*aa*cc;
          if (D >= 0)
             K_c(1) = (-bb + sqrt(D))/(2*aa);
             K_c(2) = (-bb - sqrt(D))/(2*aa);
          end;
       elseif (bb ~= 0)
          K_c(1) = -cc/bb;
       end;
       for i=1:2
          // check 3rd root
          f = a3 + K_c(i)*b3;
          if (f<-1 | f>=1)
             K_c(i) = 0;
          else
             // check real parts of ultimate roots
             alfa(i) = (a1-a3 + K_c(i)*(b1-b3))/(-2);
             if ((alfa(i)<-1) | (alfa(i)>=1))

                K_c(i) = 0;//can't compute gain for complex roots
             end;
          end;
       end;
    
       K_r=0;	//gain for real roots // Defined by the user    
    
       if (-b1+b2-b3 ~= 0)
          K_r = (-1+a1-a2+a3)/(-b1+b2-b3);
          bb = a2-a3 + K_r*(b2-b3);

          cc = a3 + K_r*b3;

          // check othes 2 roots are stable
          f1 = (-bb + sqrt(bb*bb-4*cc))/2;
          f2 = (-bb - sqrt(bb*bb-4*cc))/2;
          if (abs(f1)>1 | f1==1 | abs(f2)>1 | f2==1)
             K_r = 0;
          end;
       end
    
       //Use 2 complex roots if possible
       if (K_c(1) > 0 | (K_c(1)<0 & K_c(2)<=0 & K_r<=0))
          Kpu = K_c(1);
          Tu = T0 * 2*%pi/acos(alfa(1));
       elseif (K_c(2) > 0 | (K_c(2)<0 & K_r<=0))
          Kpu = K_c(2);
          Tu = T0 * 2*%pi/acos(alfa(2));
       elseif (K_r~=0)
          Kpu = K_r;
          Tu = T0 * 2;
       else
          //default values
          Kpu = 1; 
          Tu = 2*T0;
       end;

        TI = 0.5*Tu;
        TD = 0.125*Tu;
        Kp = 0.6*Kpu;
        q0 = Kp*(1+TD/T0);
        q1 = -Kp*(1-T0/TI+2*TD/T0);
        q2 = Kp*TD/T0;
        
	// PID
        if k>4 then
            u(k-1) = q0*e(k-1) + q1*e(k-2) + q2*e(k-3) + u(k-2)
        else
            u(k-1) =  q0*e(k-1)  
        end

///////////////////////////////////////////////////////////////////////////////
//  Single Neuron PID controller  

    elseif method == 'Single Neuron' | method == '7' then    

        // Adjustable parameters
        kap = 0.9;    // Defined by the user    
        mu = [1 1 1]  // Defined by the user    
        
        x1 = r(k-1)-y(k-1);
        x2 = x1-X(k-1,1);
        x3 = x2-X(k-1,2);
        
        X(k,:) = [x1 x2 x3];
        
        if k==kk then
            if i(k-1)==0 then
                var=sign(y(k-1));
            else
                var=sign(y(k-1)/i(k-1));
            end
        else
            if i(k-1)-i(k-2)==0 then
                var=sign(y(k-1)-y(k-2));
            else
                var=sign((y(k-1)-y(k-2))/(i(k-1)-i(k-2)));
            end
        end
        w1 = w(k-1,1) + mu(1)*kap*x1*X(k-1,1)*var
        w2 = w(k-1,2) + mu(2)*kap*x1*X(k-1,2)*var
        w3 = w(k-1,3) + mu(3)*kap*x1*X(k-1,3)*var
        
        w(k,:) = [w1 w2 w3]; 
        
        s_w = abs(w1)+abs(w2)+abs(w3)
        n_w1 = w1/s_w
        n_w2 = w2/s_w
        n_w3 = w3/s_w
        
        i(k) = i(k-1) + kap*(n_w1*x1 + n_w2*x2 + n_w3*x3)
        
        kp = kap*n_w1
        ki = kap*n_w2
        kd = kap*n_w3
        theta(1:3,k-1)=[kp;ki;kd]
        
        if k>2 then
            u(k-1) = u(k-2) + kp*x1 + ki*x2 + kd*x3;
        else
            u(k-1) =  kp*x1 + ki*x2 + kd*x3;
        end
        
///////////////////////////////////////////////////////////////////////////////
//  Least Squares with Suport Vectro Regression and Kernel Tuning 

    elseif method == 'LSSVR' | method == '8' |method == 'LSSVRKT' | method == '9' then 

        //////////////////////Adjustable Variables
        eta = 0.9; // learning rate 0<eta<1 - The bigger the faster // Defined by the user    
        C = 320 // Defined by the user    
        L = 30;     //size of the sliding window // Defined by the user    
        nu = L;    //number of u elements in the support vector // Defined by the user    
        ny = L;    //number of y elements in the support vector // Defined by the user    
        //////////////////////////////////////////
        if k==kk then
            sig(k-1) = 0.9//
            vx(1:nu+ny,k-1) = zeros(nu+ny,1)
            I_mU = eye(L,L)
            xc1 = e(k-1) - e(k-2)
            xc2 = e(k-1)
            xc3 = e(k-1) - 2*e(k-2)  
        else
            xc1 = e(k-1) - e(k-2)
            xc2 = e(k-1)
            xc3 = e(k-1) - 2*e(k-2) + e(k-3) 
        end
        
        kp = theta(1,max(k-2,1))
        ki = theta(2,max(k-2,1))
        kd = theta(3,max(k-2,1))
        

        vu = [u(k-2:-1:max(k-L-1,1));zeros(2+L-k,1)]
        vy = [y(k-1:-1:max(k-L,1));zeros(1+L-k,1)]
       vx=[[vu;vy] vx(:,1:min(k-1,L-1))]
        
        An = I_mU(1:L-1,1:L-1)

        mU=inv(I_mU)
        
        q = kapa(vx(:,1),vx(:,1),sig(k-1))
        for j=1:L-1
            if j<k then
                Q(j) = kapa(vx(:,1),vx(:,j+1),sig(k-1))'
            else
                Q(j) = 0    
            end
        end
        
        I_mU=[q Q';Q An]+eye(L,L)*(1/C); 

        ones_v = ones(1,L) 
        b = ones_v*mU*vy/(ones_v*mU*ones_v')
        alpha = mU*(vy-ones_v'*b)
        
        ye(k-1)=b
        for j=1:L-1
           if j<k then 
                  ye(k-1) = ye(k-1) + alpha(j,1)*kapa(vx(:,1),vx(:,j+1),sig(k-1))
           end
        end  
        ep(k-1) = y(k-1) - ye(k-1)
        
        dye_du = 0;
        for j=1:L-1
            if j<k then
                dye_du = dye_du - alpha(j,1)*(vy(1,1)-vy(j+1,1))*kapa(vx(:,1),vx(:,j+1),sig(k-1))/(sig(k-1)^2)
            end
        end
            
        kp = max( kp - eta*xc2*dye_du*xc1,0)
        ki = max( ki - xc2*eta*xc2*dye_du*xc2,0)
        kd = max( kd - eta*xc2*dye_du*xc3,0)
        theta(1:3,k-1)=[kp ki kd]'
        
        if k>2 then
            u(k-1) = u(k-2) + kp*xc1 + ki*xc2 + kd*xc3
        else
            u(k-1) =  kp*xc1 + ki*xc2 + kd*xc3
        end        

        if method == 'LSSVRKT' | method == '9' then
            dye_ds = 0;
            for j=1:L-1
                if j<k then
                    dye_ds = dye_ds + alpha(j)*kapa(vx(:,1),vx(:,j+1),sig(k-1))*(vx(:,1)-vx(:,j+1))'*(vx(:,1)-vx(:,j+1))/(sig(k-1)^3)                    
                end
            end
            sig(k) = sig(k-1) + eta*ep(k-1)*dye_ds
        else
            sig(k)=sig(k-1)
        end
    end
    
    if Max_u_var ~= -1 then
        if (u(k-1)-u(k-2))>Max_u_var then
            u(k-1) = u(k-2) + Max_u_var;
        elseif (u(k-1)-u(k-2))<-Max_u_var then
            u(k-1) = u(k-2) - Max_u_var;
        end
    end
    
    if Max_u ~=-1 then
        if u(k-1) > Max_u then
            u(k-1) = Max_u;
        elseif u(k-1) < Min_u then
            u(k-1) = Min_u;
        end
    end
endfunction

function ret=kapa(xi,xj,sig)
    ret = exp(-norm((xi-xj))^2/(2*sig*sig))
endfunction
